// Industry Standard algorithm
//

#include "stdafx.h"
#include <stack>

using namespace std;
using namespace boost::heap;

typedef binomial_heap<Label>::handle_type HeapHandle;

class IS {
private:

void addNewLabel( Label labelNew) { 
	if(nodeLabelsNoBreak[labelNew.node] != -1){
		Label c = allLabels[nodeLabelsNoBreak[labelNew.node]];
		if(!c.isRemoved){
			pq.removeLabel_nobreak(c.id);
			c.isRemoved = true;
		}
	}
	nodeLabelsNoBreak[labelNew.node] = labelNew.id;
	pq.addLabel_nobreak(labelNew.id, labelNew.arrivalTime);
	allLabels.push_back(labelNew);
}


public:
	static const bool ASTAR = false;
	int BREAK_TIME; // 45 minutes
	int THRESHOLD; // 4.5 hour
	int TESTCASE_NUMBER;
	int START_NODE;
	int DESTINATION_NODE;
	int DEPARTURE_TIME;
	int SHIFT_START;
	vector< int > nodeLabelsNoBreak;
	vector< int > nodeLabelsBreak;
	vector< Label > allLabels;
	SearchPQ pq;
	NetworkData network;

	IS(int t, NetworkData n, int s, int d, int dtime, int shifttime ) {
		TESTCASE_NUMBER = t;
		network = n;
		START_NODE = s;
		DESTINATION_NODE = d;
		DEPARTURE_TIME = dtime;
		SHIFT_START = shifttime;

		nodeLabelsNoBreak.resize(network.getNumberOfNodes());
		fill(nodeLabelsNoBreak.begin(), nodeLabelsNoBreak.end(), -1);
		nodeLabelsBreak.resize(network.getNumberOfNodes());
		fill(nodeLabelsBreak.begin(), nodeLabelsBreak.end(), -1);

	}
	~IS() {  
		nodeLabelsNoBreak.clear();
		nodeLabelsBreak.clear();
		allLabels.clear();
		pq.clear();
	}  
int search(std::string results_file, bool longBreaks){
	pq.clear();
	if(longBreaks){
		BREAK_TIME = 39600;
		THRESHOLD = 32400;
	} else{
		BREAK_TIME = 2700;
		THRESHOLD = 16200;
	}

	std::clock_t start;
	start = std::clock();
	
	Label labelStart(0, START_NODE, DEPARTURE_TIME, SHIFT_START, false, 0, -1);
	pq.addLabel_nobreak(0,DEPARTURE_TIME);
	nodeLabelsNoBreak[START_NODE] = 0;
	allLabels.push_back(labelStart);

	bool stop = false;
	int nextLabelID = 1;

	//STEP 1: no-break labels + explore break labels
	int arrival_time = std::numeric_limits<int>::max();

	while( !pq.isEmpty_nobreak() && !stop){
		int currentLabelId = 0;
		currentLabelId = pq.getFirstLabel_nobreak();

		pq.deleteFirstLabel_nobreak();
		allLabels[currentLabelId].isRemoved = true;
		Label current =  allLabels[currentLabelId];
		
		if(current.node == DESTINATION_NODE){
			stop = true;
		}		
		
		int beginEdge = network.getFirstEdgeOfNode(current.node);
		int endEdge = network.getEndEdgeOfNode(current.node);

		for ( int i = beginEdge; i != endEdge; i++ )
		{
			int naarNode = network.getNaarNode(i);
			std::tuple<int,int,int> block = network.blockedTime(i, current.arrivalTime); //start,end,blocktime
			double speed = network.getConstantSpeed(i); // speed converted from km/h to m/s

			//get length of link
			double length = network.getLength(i); // length in meters
			//calculate f(tau)
			int traveltime = std::get<2>(block) + max(1, int((((36 * length) / speed+5) / 10))); //seconds 

			int shifttime =  max(1, int((((36 * length) / speed+5) / 10)));
		
			Label currentNaar; bool naar = false;
			if( nodeLabelsNoBreak[naarNode] != -1){
				currentNaar = allLabels[nodeLabelsNoBreak[naarNode]];
				naar = true;
			}

			if( !naar || current.arrivalTime + traveltime <  currentNaar.arrivalTime || (current.arrivalTime + traveltime == currentNaar.arrivalTime && current.shiftDuration + shifttime < currentNaar.shiftDuration)){
				if(current.shiftDuration + shifttime < 2 * THRESHOLD ) {
					Label labelNew(nextLabelID, naarNode, current.arrivalTime+traveltime, current.shiftDuration + shifttime, false, 0, current.id);
					addNewLabel( labelNew );
					nextLabelID++;
				}
			}
		}
	}

	if ( nodeLabelsNoBreak[DESTINATION_NODE] != -1 ) {
		if ( allLabels[nodeLabelsNoBreak[DESTINATION_NODE]].shiftDuration <= THRESHOLD ) {
			//writePath(allLabels[nodeLabelsNoBreak[DESTINATION_NODE]],  allLabels, results_file);
			std::ofstream results;
			results.open(results_file, std::ofstream::out | std::ofstream::app);
			results << TESTCASE_NUMBER <<" " << normalizedTime(allLabels[nodeLabelsNoBreak[DESTINATION_NODE]].arrivalTime) << " " << 0 << " 0 " <<((double)(std::clock() - start)/1000) << " 0" << std::endl;
			return normalizedTime(allLabels[nodeLabelsNoBreak[DESTINATION_NODE]].arrivalTime);
		}
		else {
			Label current = allLabels[nodeLabelsNoBreak[DESTINATION_NODE]];
			std::stack<int> path = stack<int>();
			path.push(current.id);
			while(current.shiftDuration >= THRESHOLD ) {
				current = allLabels[current.prec_id];
				path.push(current.id);
			}
			//break is taken between prev and next (==path.top())
			
			Label prev = allLabels[path.top()];  path.pop();
			Label next = allLabels[path.top()];  path.pop();

			// take a break ... between prev and next
			int naarNode = next.node;
			int edge = network.getEdge(prev.node, next.node);
			std::tuple<int,int,int> block = network.blockedTime(edge, prev.arrivalTime); //start,end,blocktime
			double speed = network.getConstantSpeed(edge); // speed converted from km/h to m/s
			double length = network.getLength(edge); // length in meters
			int traveltime = std::get<2>(block) + max(1, int((((36 * length) / speed+5) / 10))); //seconds 
			int shifttime =  max(1, int((((36 * length) / speed+5) / 10)));
			int travelBreaktime = traveltime + BREAK_TIME;
			int shiftBreaktime = 0; 
			if(THRESHOLD - current.shiftDuration >= shifttime){ 
				shiftBreaktime = 0; 
			}
			else{
				shiftBreaktime = shifttime - ( THRESHOLD - current.shiftDuration );
			}
			if(std::get<0>(block) != std::get<1>(block)){
				if(current.arrivalTime >= std::get<0>(block) - traveltime && current.arrivalTime < std::get<1>(block) - BREAK_TIME){
					travelBreaktime = traveltime;
					shiftBreaktime = shifttime;
				}
				if(current.arrivalTime >= std::get<1>(block) - BREAK_TIME && current.arrivalTime < std::get<1>(block)) {
					shiftBreaktime = shifttime;
				} 
			}
			next.arrivalTime = prev.arrivalTime + travelBreaktime;
			next.shiftDuration = shiftBreaktime;
			next.breakTaken = true;
			next.breakNode = prev.node;

			prev = next;
			// calculate new arrival time at node
			while ( ! path.empty() ) {
				next = allLabels[path.top()];	path.pop();
				int naarNode = next.node;
				int edge = network.getEdge(prev.node, next.node);
				std::tuple<int,int,int> block = network.blockedTime(edge, prev.arrivalTime); //start,end,blocktime
				double speed = network.getConstantSpeed(edge); // speed converted from km/h to m/s
				double length = network.getLength(edge); // length in meters
				int traveltime = std::get<2>(block) + max(1, int((((36 * length) / speed+5) / 10))); //seconds 
				int shifttime =  max(1, int((((36 * length) / speed+5) / 10)));
				next.arrivalTime = prev.arrivalTime + traveltime;
				next.shiftDuration = prev.shiftDuration + shifttime;
				next.breakTaken = prev.breakTaken;
				next.breakNode = prev.breakNode;
				prev = next;
			}

			if(next.shiftDuration > THRESHOLD ) {
				//writePath(allLabels[nodeLabelsNoBreak[DESTINATION_NODE]],  allLabels, results_file);
				std::ofstream results;
				results.open(results_file, std::ofstream::out | std::ofstream::app);
				results << TESTCASE_NUMBER <<" " << std::numeric_limits<int>::max() << " " << 0 << " 0 " <<((double)(std::clock() - start)/1000) << " 0" << std::endl;
				return std::numeric_limits<int>::max();
			}
			else {
				//writePath(allLabels[nodeLabelsNoBreak[DESTINATION_NODE]],  allLabels, results_file);
				std::ofstream results;
				results.open(results_file, std::ofstream::out | std::ofstream::app);
				results << TESTCASE_NUMBER <<" " << normalizedTime(next.arrivalTime) << " " << next.breakTaken << " " << next.breakNode << " " << current.arrivalTime  << " " << ((double)(std::clock() - start)/1000) << std::endl;
				//results << TESTCASE_NUMBER <<" " << normalizedTime(next.arrivalTime) << " " << next.breakTaken << " "<< next.breakNode << " "  << ((double)(std::clock() - start)/1000) << " 0" <<std::endl;

				return normalizedTime(next.arrivalTime);
			}
			
		}


	}
	else {
		std::ofstream results;
		results.open(results_file, std::ofstream::out | std::ofstream::app);
		results << TESTCASE_NUMBER <<" " << std::numeric_limits<int>::max() << " " << 0 << " " << 0 << " " <<((double)(std::clock() - start)/1000) << " 0" << std::endl;

		return std::numeric_limits<int>::max();
	}

}

};