#include "stdafx.h"
#include <fstream>
#include <iostream>
#include <string>
#include <vector>
#include <map>
#include <array>
#include <algorithm>
#include <numeric>
#include <cstdio>
#include <ctime>
#include <cmath>
#include <tuple>
#include <cstdint>
#include <assert.h>
using namespace std;

//NetworkData reads the map data (provided by HERE)
class NetworkData
{
private:
	std::vector<int> nodes;
	std::vector<int> firstList; 
	std::vector<int> naarList;
	std::vector<int> lengthList;
	std::vector<int> avgspeeds;
	std::vector<int> country;
	std::vector<int> roadtype;
	std::vector<int> unconnected;
	std::map< int, std::vector < std::array<int,2> > > roadBlocks;
	std::vector<int> coordinates; //first x coordinates of all nodes, then y coordinates of all nodes

	std::vector<int> geoDistances;

	void readFirst(){
		ifstream first(mapFolder + "first.net", std::ios::binary);

		//find last char in stream
		first.seekg(0,first.end);
		int length = first.tellg();

		//go back to begin of stream
		first.seekg(0,first.beg);
		numberOfNodes = length/4 - 1; //-1 since last node represents the last position of the file. This will be saved though

		firstList.resize(length/4);
		first.read(reinterpret_cast<char*>(&firstList[0]), length*4);

		numberOfEdges = firstList[numberOfNodes];
	}

	void readNodes(){
		ifstream orgnr(mapFolder+"org_nr.net", std::ios::binary);	
		nodes.resize(numberOfNodes);
		orgnr.read(reinterpret_cast<char*>(&nodes[0]), numberOfNodes*4);
	}

	void readNaar(){
		ifstream naar(mapFolder+"naar.net", std::ios::binary);	
		naarList.resize(numberOfEdges);
		naar.read(reinterpret_cast<char*>(&naarList[0]), numberOfEdges*4);
	}

	void readLength(){
		ifstream length(mapFolder+"lengtelong.net", std::ios::binary);	
		lengthList.resize(numberOfEdges);
		length.read(reinterpret_cast<char*>(&lengthList[0]), numberOfEdges*4);
	}

	void readAverageSpeed(){
		ifstream type(mapFolder+"type.net", std::ios::binary);
		avgspeeds.resize(numberOfEdges);
		roadtype.resize(numberOfEdges);
		for(int i=0; i!=numberOfEdges; i++){
			type.read(reinterpret_cast<char*>(&roadtype[i]),1);
		}
		int speedProfile[19] = { 78, 75, 69, 65, 56, 54, 47, 45, 36, 33, 24, 20, 15, 11, 9, 7, 4, 1, 2 };
		transform(begin(roadtype), end(roadtype), begin(avgspeeds), [&speedProfile](int elem){ return speedProfile[(int) elem]; });
	}

	void readCountry(){
		ifstream lnkcou(mapFolder+"lnkcou.net", std::ios::binary);	
		country.resize(numberOfEdges);
		for(int i = 0; i!= numberOfEdges; i++){
			lnkcou.read(reinterpret_cast<char*>(&country[i]), 2);
		}
	}

	void readCoordinates(){
		ifstream xypos(mapFolder+"xypos.net",std::ios::binary);
		coordinates.resize(numberOfNodes*2);
		xypos.read(reinterpret_cast<char*>(&coordinates[0]), numberOfNodes*4*2);
	}

	void readRoadBlocks(){
		ifstream blocks(mapFolder+"roadBlocks.net",std::ios::binary);
		int countrycode = 0;
		int elements = 56;

		for(int i = 0; i!= elements; i++){
			std::array<int,2> blockTime;
			blocks.read(reinterpret_cast<char*>(&countrycode),2);
			blocks.read(reinterpret_cast<char*>(&blockTime[0]),4);
			blocks.read(reinterpret_cast<char*>(&blockTime[1]),4);
			blockTime[0] = (int)blockTime[0]/10;
			blockTime[1] = (int)blockTime[1]/10;
			if(blockTime[0] !=  blockTime[1]) {
				roadBlocks[countrycode].push_back(blockTime);
			}
		}
	}

	void readUnConnected(){
		ifstream uc(mapFolder+"unconnected.net", std::ios::binary);	
		unconnected.resize(numberOfNodes);
		uc.read(reinterpret_cast<char*>(&unconnected[0]), numberOfNodes*4);
	}


public:
	std::string mapFolder;
	int numberOfNodes;
	int numberOfEdges;

	~NetworkData(void){
		nodes.clear();
		firstList.clear();
		naarList.clear();
		lengthList.clear();
		avgspeeds.clear();
		country.clear();
		roadtype.clear();
		unconnected.clear();
		roadBlocks.clear();
		coordinates.clear();
		geoDistances.clear();
	};
	NetworkData(void){};
	NetworkData(std::string folder, bool astar){
		std::clock_t start;
		std::cout << "Reading network data ... " ;
		double duration;
		start = std::clock();
		
		mapFolder = folder;
		readFirst();
		readNodes();
		readNaar();
		readLength();
		readAverageSpeed();
		readCountry();	

		readRoadBlocks();

		if(astar){
			readCoordinates();
		}
		duration = ( std::clock() - start ) / (double) CLOCKS_PER_SEC;
		std::cout<<"took: "<< duration << " seconds" << std::endl;
		std::cout<<"Network has " << getNumberOfNodes() << " nodes and " << getNumberOfEdges() << " edges" << std::endl;
	}
	
	int getNumberOfEdges() const { return numberOfEdges; }
	int getNumberOfNodes() const { return numberOfNodes; }
	int getNaarNode(int id) const { return naarList[id]; }
	int getLength(int edgeid) const {return lengthList[edgeid]; }

	int getFirstEdgeOfNode ( int nodeid ) {
		return firstList[nodeid];
	}

	int getEndEdgeOfNode ( int nodeid ) {
		return firstList[nodeid+1];
	}

	int getNodeId( int order) {
		return nodes[order];
	}

	int getIdOfNode( int nodeid ) {
		for(int i = 0; i < getNumberOfNodes(); i++) {
			if(nodes[i] == nodeid ) {
				return i;
			}
		}
		return -1;
	}

	int getEdge ( int start, int dest ) {
		int startEdge = getFirstEdgeOfNode(start);
		int endEdge = getEndEdgeOfNode(start);
		for(int i = startEdge; i< endEdge; i++ ) {
			if (getNaarNode(i) == dest ) {
				return i;
			}
		}
		return -1;

	}

	int getConstantSpeed(int edge) const{
		return avgspeeds[edge];
	}

	int normalizedTimeX(int time) const {
		if(time >= 604800)
			return time-604800;
		if(time < 0)
			return time + 604800;
		return time;
	}

	std::tuple<int,int,int> blockedTime(int edge, int departure_time) const{
		int countryEdge = country[edge];
		int default_travel_time = max(1, int((((36 * getLength(edge)) / getConstantSpeed(edge)+5) / 10)));
		departure_time = departure_time %  604800; 
		try{
			std::vector < std::array<int,2> > countryBlocks = roadBlocks.at(countryEdge);

			int start_time = (*(countryBlocks.end()-1))[0] ;
			int duration = normalizedTimeX((*(countryBlocks.end()-1))[1] - start_time);
			if(departure_time < start_time + duration - 604800){
				return std::make_tuple((*(countryBlocks.end()-1))[0], (*(countryBlocks.end()-1))[1], start_time + duration - 604800 - departure_time);		
			}
			for(auto it = countryBlocks.begin() ; it != countryBlocks.end(); it++){
				std::array<int,2> block = *it;
				start_time = block[0] ;
				duration = normalizedTimeX(block[1] - block[0]);
				if ( departure_time <= block[0] -  default_travel_time)
					return std::make_tuple(0, 0, 0);
				if ( departure_time <= block[0] ){
					return std::make_tuple(block[0], block[1], duration);
				} if ( departure_time <= block[1] ){
					return std::make_tuple(block[0], block[1], start_time + duration - departure_time);		
				}
			}
			start_time = (*countryBlocks.begin())[0];
			duration = normalizedTimeX((*countryBlocks.begin())[1] - start_time);
			if ( departure_time > start_time + 604800 - default_travel_time ){
				return std::make_tuple((*countryBlocks.begin())[0], (*countryBlocks.begin())[1], duration);
			}

		} catch(const std::out_of_range& oor) {
			//non existing key in map
		}
		return std::make_tuple(0,0,0);
	}


	double getX(int node) const{
		assert(node >=0 && node < getNumberOfNodes() );
		return coordinates[node];
	}
	
	double getY(int node) const{
		assert(node >=0 && node < getNumberOfNodes() );
		return coordinates[numberOfNodes + node];
	}

	double toRadians(double deg){
		return (deg * atan(1)*4) / 180;
	}

	double calcDistance(int node1, int node2){
		double lat1 = getY(node1)/10000;
		double lat2 = getY(node2)/10000;
		double long1 = getX(node1)/10000; 
		double long2 = getX(node2)/10000;

		double dlat = toRadians(abs(lat1-lat2));
		double dlong = toRadians(abs(long1-long2));

		double a = sin(dlat/2) * sin(dlat/2) + cos(toRadians(lat1))* cos(toRadians(lat2)) * sin(dlong/2) * sin(dlong/2);
		double c = 2 * atan2(sqrt(a),sqrt(1-a));
		double distance = c * 6371000 / 1000;
		return (int)distance; //in km

	}

	std::vector<int> calcTimes(int destination){
		std::vector<int> time;
		time.resize(numberOfNodes);
		for(int i=0; i != numberOfNodes; i++){
			time[i] = (int) (calcDistance(i, destination) / 80)*60*60*10; //80 km/h estimate ==hour. *60 == minutes *60==sec *10
		}
		return time;
	}

	int getCountry(int nodeid){
		return country[getFirstEdgeOfNode(nodeid)];
	}
};