// SBG: Pareto multilabel search (optimal break planning)
//

#include <unordered_map>

using namespace std;
using namespace boost::heap;

typedef binomial_heap<Label>::handle_type HeapHandle;

class SBG {
private:

bool isDominated( Label & labelNew ){
	int i = 0;
	auto it = nodeLabels.find(labelNew.node);
	if(it != nodeLabels.end()){
		vector<int> & labels = it->second;
		while(i < labels.size() ) {
			if(labelNew.breakTaken == allLabels[ labels[i]].breakTaken && (labelNew.arrivalTime) >= allLabels[ labels[i]].arrivalTime && labelNew.shiftDuration >= allLabels[labels[i]].shiftDuration){
				return true;
			}
			i++;
		}
	}
	return false;
}

void removeDominated( Label & labelNew ){
	int i = 0;
	auto it = nodeLabels.find(labelNew.node);
	if(it != nodeLabels.end()){
		vector<int> & labels = it->second;
		while(i < labels.size() ) {
			if(labelNew.breakTaken ==  allLabels[labels[i]].breakTaken && (labelNew.arrivalTime) <= allLabels[labels[i]].arrivalTime && labelNew.shiftDuration <= allLabels[labels[i]].shiftDuration){
				if((labelNew.arrivalTime) < allLabels[ labels[i]].arrivalTime || labelNew.shiftDuration < allLabels[labels[i]].shiftDuration){
					Label c = allLabels[labels[i]];
					if(!c.isRemoved) {
						pq.removeLabel_nobreak(c.id);
						c.isRemoved = true;
					}
					iter_swap(labels.begin()+i, labels.end()-1);
					labels.pop_back();
				} else {
					i++;
				}
			} else{
				i++;
			}
		}
	}
}

void addNewLabel ( Label & labelNew) {
	assert(labelNew.shiftDuration < THRESHOLD);
	auto it = nodeLabels.find(labelNew.node);
	if(it == nodeLabels.end()){
		nodeLabels[labelNew.node].push_back(labelNew.id);
		pq.addLabel_nobreak(labelNew.id, labelNew.arrivalTime);
		allLabels.push_back(labelNew);
		nextLabelID++;
	} else{
		// add label if it is not dominated by any other label in the set
		vector<int> & labels = it->second;
		bool dominated = isDominated( labelNew );
		if(!dominated){
			//add label to set
			//remove dominated labels given the new label
			labels.push_back(labelNew.id);
			pq.addLabel_nobreak(labelNew.id, labelNew.arrivalTime);
			allLabels.push_back(labelNew);
			removeDominated( labelNew );
			nextLabelID++;
		}
	}	
}


public:
	static const bool ASTAR = false;
	int BREAK_TIME; // 11 hours
	int THRESHOLD; // 9 hours
	int TESTCASE_NUMBER;
	int START_NODE;
	int DESTINATION_NODE;
	int DEPARTURE_TIME;
	int SHIFT_START;
	unordered_map <int, vector<int> > nodeLabels;

	vector< Label > allLabels;
	SearchPQ pq;
	std::vector<int> avgTime;
	NetworkData* network;
	int nextLabelID;

	SBG(int t, NetworkData & n, int s, int d, int dtime, int shifttime ) {
		TESTCASE_NUMBER = t;
		network = &n;
		START_NODE = s;
		DESTINATION_NODE = d;
		DEPARTURE_TIME = dtime;
		SHIFT_START = shifttime;

	}

int search(std::string results_file, bool longBreaks){
	if(longBreaks){
		BREAK_TIME = 39600;
		THRESHOLD = 32400;
	} else{
		BREAK_TIME = 2700;
		THRESHOLD = 16200;
	}
	std::clock_t start;
	start = std::clock();
	
	Label labelStart(0, START_NODE, DEPARTURE_TIME, SHIFT_START, false, 0, -1);
	pq.addLabel_nobreak(0,DEPARTURE_TIME);
	nodeLabels[START_NODE].push_back(0);
	allLabels.push_back(labelStart);

	nextLabelID = 1;
	bool stop = false;

	int arrival_time = std::numeric_limits<int>::max();
	while( !pq.isEmpty_nobreak() && !stop){

		if(pq.getPriorityFirst_nobreak() > arrival_time) {
			stop = true;
		}

		int currentLabelId = 0;
		currentLabelId = pq.getFirstLabel_nobreak();
		int priority = pq.getPriorityFirst_nobreak();
		
		pq.deleteFirstLabel_nobreak();
		allLabels[currentLabelId].isRemoved = true;
		Label current =  allLabels[currentLabelId];
		if(current.node == DESTINATION_NODE) { 
			arrival_time = current.arrivalTime;
		}
		
		int beginEdge = network->getFirstEdgeOfNode(current.node);
		int endEdge = network->getEndEdgeOfNode(current.node);

		for ( int i = beginEdge; i != endEdge; i++ )
		{
			int naarNode = network->getNaarNode(i);
			std::tuple<int,int,int> block = network->blockedTime(i, current.arrivalTime); //start,end,blocktime
			double speed = network->getConstantSpeed(i); // speed converted from km/h to m/s

			//get length of link
			double length = network->getLength(i); // length in meters
			//calculate f(tau)
			int traveltime = std::get<2>(block) + max(1, int((((36 * length) / speed+5) / 10))); //seconds 
			int shifttime =  max(1, int((((36 * length) / speed+5) / 10)));

			int travelBreaktime = traveltime + BREAK_TIME;
			int shiftBreaktime = 0; 
			if(THRESHOLD - current.shiftDuration >= shifttime){ 
				shiftBreaktime = 0; 
			}
			else{
				shiftBreaktime = shifttime - ( THRESHOLD - current.shiftDuration );
			}
			if(std::get<0>(block) != std::get<1>(block)){
				if(current.arrivalTime >= std::get<0>(block) - traveltime && current.arrivalTime < std::get<1>(block) - BREAK_TIME){
					travelBreaktime = traveltime;
					shiftBreaktime = shifttime;
				}
				if(current.arrivalTime >= std::get<1>(block) - BREAK_TIME && current.arrivalTime < std::get<1>(block)) {
					travelBreaktime = traveltime + BREAK_TIME;
					shiftBreaktime = shifttime;
				} 
			}
			
			if(current.shiftDuration + shifttime < THRESHOLD) {
				Label labelNew(nextLabelID, naarNode, current.arrivalTime+traveltime, current.shiftDuration + shifttime, current.breakTaken, current.breakNode, current.id);
				addNewLabel( labelNew );
			}
			if(!current.breakTaken && shiftBreaktime < THRESHOLD) {
				Label labelNewB(nextLabelID, naarNode, current.arrivalTime + travelBreaktime, shiftBreaktime, true, current.node, current.id);
				addNewLabel( labelNewB );
			}
		}
	}

	// STEP 2 : Get best result of break and non-break options
	int xarrival_time = std::numeric_limits<int>::max();
	bool breaktaken = false;
	int shift_duration = std::numeric_limits<int>::max();
	int breaknode = 0;
	Label lab;
	auto it = nodeLabels.find(DESTINATION_NODE);
	if ( it != nodeLabels.end() ) {
		std::vector<int> & labels = it-> second;
		for(int i = 0; i < labels.size() ; i++ ) {
			if(allLabels[labels[i]].arrivalTime < xarrival_time ){
				xarrival_time = allLabels[labels[i]].arrivalTime;
				breaktaken = allLabels[labels[i]].breakTaken;
				breaknode = allLabels[labels[i]].breakNode;
				shift_duration = allLabels[labels[i]].shiftDuration;
				lab = allLabels[labels[i]];
				//writePath(allLabels[nodeLabelsBreak[DESTINATION_NODE][i]],  allLabels, "D:/marieket/Documents/Maps/NBNL4DEGB3Eur1C09/Results/path1-multilabel.txt");
			}
		}
	}
	if( xarrival_time != std::numeric_limits<int>::max() ) {
		xarrival_time = normalizedTime(xarrival_time);
	}	
	std::ofstream results;
	results.open(results_file, std::ofstream::out | std::ofstream::app);
	results << TESTCASE_NUMBER <<" " << xarrival_time << " " << breaktaken << " "<< breaknode << " " <<((double)(std::clock() - start)/1000) << " 0" <<std::endl;
	writePath(lab,  allLabels, results_file);
	std::cout << "noparking: " << (double)network->getY(breaknode )/10000 << "," << (double)network->getX(breaknode )/10000 << std::endl;

	return xarrival_time;
	}
};