// SBG-SL: Single label search (heuristic break planning)


#include "stdafx.h"

using namespace std;
using namespace boost::heap;

typedef binomial_heap<Label>::handle_type HeapHandle;

class SBGSL {
private:

void addNewLabel( Label labelNew) { 
	if(labelNew.breakTaken){
		if(nodeLabelsBreak[labelNew.node] != -1){
			Label c = allLabels[nodeLabelsBreak[labelNew.node]];
			if(!c.isRemoved){
				pq.removeLabel_break(c.id);
				c.isRemoved = true;
			}
		}
		nodeLabelsBreak[labelNew.node] = labelNew.id;
		pq.addLabel_break(labelNew.id,labelNew.shiftDuration );
		allLabels.push_back(labelNew);
	} else {
		if(nodeLabelsNoBreak[labelNew.node] != -1){
			Label c = allLabels[nodeLabelsNoBreak[labelNew.node]];
			if(!c.isRemoved){
				pq.removeLabel_nobreak(c.id);
				c.isRemoved = true;
			}
		}
		nodeLabelsNoBreak[labelNew.node] = labelNew.id;
		pq.addLabel_nobreak(labelNew.id, labelNew.arrivalTime);
		allLabels.push_back(labelNew);
	}
}


public:
	static const bool ASTAR = false;
	int BREAK_TIME; 
	int THRESHOLD;
	int TESTCASE_NUMBER;
	int START_NODE;
	int DESTINATION_NODE;
	int DEPARTURE_TIME;
	int SHIFT_START;
	vector< int > nodeLabelsNoBreak;
	vector< int > nodeLabelsBreak;
	vector< Label > allLabels;
	SearchPQ pq;
	NetworkData network;

	SBGSL(int t, NetworkData n, int s, int d, int dtime, int shifttime ) {
		TESTCASE_NUMBER = t;
		network = n;
		START_NODE = s;
		DESTINATION_NODE = d;
		DEPARTURE_TIME = dtime;
		SHIFT_START = shifttime;

		nodeLabelsNoBreak.resize(network.getNumberOfNodes());
		fill(nodeLabelsNoBreak.begin(), nodeLabelsNoBreak.end(), -1);
		nodeLabelsBreak.resize(network.getNumberOfNodes());
		fill(nodeLabelsBreak.begin(), nodeLabelsBreak.end(), -1);
	}
	~SBGSL() {  
		nodeLabelsNoBreak.clear();
		nodeLabelsBreak.clear();
		allLabels.clear();
		pq.clear();
	}  


int search(std::string results_file, bool longBreaks){
	pq.clear();
	if(longBreaks){
		BREAK_TIME = 39600;
		THRESHOLD = 32400;
	} else{
		BREAK_TIME = 2700;
		THRESHOLD = 16200;
	}
	std::set<int> node_errors = std::set<int>();

	std::clock_t start;
	start = std::clock();
	
	Label labelStart(0, START_NODE, DEPARTURE_TIME, SHIFT_START, false, 0, -1);
	pq.addLabel_nobreak(0,DEPARTURE_TIME);
	nodeLabelsNoBreak[START_NODE] = 0;
	allLabels.push_back(labelStart);

	bool stop = false;
	int nextLabelID = 1;

	//STEP 1: no-break labels + explore break labels
	int arrival_time = std::numeric_limits<int>::max();
	while( !pq.isEmpty_nobreak() && !stop){

		if(pq.getPriorityFirst_nobreak() > arrival_time) {
			stop = true;
		}

		int currentLabelId = 0;
		currentLabelId = pq.getFirstLabel_nobreak();
		int priority = pq.getPriorityFirst_nobreak();



		pq.deleteFirstLabel_nobreak();
		allLabels[currentLabelId].isRemoved = true;
		Label current =  allLabels[currentLabelId];
		if(current.node == DESTINATION_NODE) { 
			arrival_time = current.arrivalTime;
		}
		
		int beginEdge = network.getFirstEdgeOfNode(current.node);
		int endEdge = network.getEndEdgeOfNode(current.node);

		for ( int i = beginEdge; i != endEdge; i++ )
		{
			int naarNode = network.getNaarNode(i);
			std::tuple<int,int,int> block = network.blockedTime(i, current.arrivalTime); //start,end,blocktime
			double speed = network.getConstantSpeed(i); // speed converted from km/h to m/s

			//get length of link
			double length = network.getLength(i); // length in meters
			//calculate f(tau)
			int traveltime = std::get<2>(block) + max(1, int((((36 * length) / speed+5) / 10))); //seconds 
			int shifttime =  max(1, int((((36 * length) / speed+5) / 10)));

			int travelBreaktime = traveltime + BREAK_TIME;
			int shiftBreaktime = 0; 
			if(THRESHOLD - current.shiftDuration >= shifttime){ 
				shiftBreaktime = 0; 
			}
			else{
				shiftBreaktime = shifttime - ( THRESHOLD - current.shiftDuration );
			}
			if(std::get<0>(block) != std::get<1>(block)){
				if(current.arrivalTime >= std::get<0>(block) - traveltime && current.arrivalTime < std::get<1>(block) - BREAK_TIME){
					travelBreaktime = traveltime;
					shiftBreaktime = shifttime;
				}
				if(current.arrivalTime >= std::get<1>(block) - BREAK_TIME && current.arrivalTime < std::get<1>(block)) {
					travelBreaktime = traveltime + BREAK_TIME;
					shiftBreaktime = shifttime;
				} 
			}
			
			Label currentNaar; bool naar = false;
			if( nodeLabelsNoBreak[naarNode] != -1){
				currentNaar = allLabels[nodeLabelsNoBreak[naarNode]];
				naar = true;
			}
			Label currentNaarBreak; bool naarB = false;
			if(nodeLabelsBreak[naarNode] != -1) {
				currentNaarBreak = allLabels[nodeLabelsBreak[naarNode]];
				naarB = true;
			}

			if( !naar || current.arrivalTime + traveltime <  currentNaar.arrivalTime || (current.arrivalTime + traveltime == currentNaar.arrivalTime && current.shiftDuration + shifttime < currentNaar.shiftDuration)){
				if(current.shiftDuration + shifttime < THRESHOLD) {
					Label labelNew(nextLabelID, naarNode, current.arrivalTime+traveltime, current.shiftDuration + shifttime, false, 0, current.id);
					addNewLabel( labelNew );
					nextLabelID++;
				}
			}

			if(naarB && (current.arrivalTime + travelBreaktime > currentNaarBreak.arrivalTime && shiftBreaktime < currentNaarBreak.shiftDuration) || (current.arrivalTime + travelBreaktime < currentNaarBreak.arrivalTime && shiftBreaktime > currentNaarBreak.shiftDuration)){
				node_errors.insert(naarNode);
			}

			if( !naarB || current.arrivalTime + travelBreaktime <  currentNaarBreak.arrivalTime || (current.arrivalTime + travelBreaktime == currentNaarBreak.arrivalTime && shiftBreaktime < currentNaar.shiftDuration)){
				if(shiftBreaktime < THRESHOLD) {
					Label labelNewB(nextLabelID, naarNode, current.arrivalTime + travelBreaktime, shiftBreaktime, true, current.node, current.id);
					addNewLabel( labelNewB );
					nextLabelID++;
				}
			}
		}
	}
	// STEP 2: 
	while( !pq.isEmpty_break()){
		int currentLabelId = 0;
		currentLabelId = pq.getFirstLabel_break();
		int priority = pq.getPriorityFirst_break();

		pq.deleteFirstLabel_break();
		allLabels[currentLabelId].isRemoved = true;
		Label current =  allLabels[currentLabelId];

		int beginEdge = network.getFirstEdgeOfNode(current.node);
		int endEdge = network.getEndEdgeOfNode(current.node);

		for ( int i = beginEdge; i != endEdge; i++ )
		{
			int naarNode = network.getNaarNode(i);
			std::tuple<int,int,int> block = network.blockedTime(i, current.arrivalTime); //start,end,blocktime
			double speed = network.getConstantSpeed(i); // speed converted from km/h to m/s
			double length = network.getLength(i); // length in meters
			
			int traveltime = std::get<2>(block) + max(1, int((((36 * length) / speed+5) / 10))); //seconds *10 =10th of second
			int shifttime =  max(1, int((((36 * length) / speed+5) / 10)));


			Label currentNaar; bool naar = false;
			if( nodeLabelsBreak[naarNode] != -1){
				currentNaar = allLabels[nodeLabelsBreak[naarNode]];
				naar = true;
			}

			if( !naar || (current.arrivalTime + traveltime <  currentNaar.arrivalTime)  || (current.arrivalTime + traveltime ==  currentNaar.arrivalTime && current.shiftDuration + shifttime < currentNaar.shiftDuration) ){
				if(current.shiftDuration + shifttime < THRESHOLD) {
					Label labelNew(nextLabelID, naarNode, current.arrivalTime+traveltime, current.shiftDuration + shifttime, true, current.breakNode, current.id);
					addNewLabel( labelNew );
					nextLabelID++;
				}
			}
			
		}
	}
	int xarrival_time = std::numeric_limits<int>::max();
	bool breaktaken = false;
	int shift_duration = std::numeric_limits<int>::max();
	int breaknode = 0; int breaktime = 0;
	if (nodeLabelsNoBreak[DESTINATION_NODE] != -1 ){
		xarrival_time = allLabels[nodeLabelsNoBreak[DESTINATION_NODE]].arrivalTime;
		breaktaken = allLabels[nodeLabelsNoBreak[DESTINATION_NODE]].breakTaken;
		breaknode = allLabels[nodeLabelsNoBreak[DESTINATION_NODE]].breakNode;
		shift_duration = allLabels[nodeLabelsNoBreak[DESTINATION_NODE]].shiftDuration;
		//writePath(allLabels[nodeLabelsNoBreak[DESTINATION_NODE]],  allLabels, results_file);
	}
	if (nodeLabelsBreak[DESTINATION_NODE] != -1 ){
		if(xarrival_time > allLabels[nodeLabelsBreak[DESTINATION_NODE]].arrivalTime){
			xarrival_time = allLabels[nodeLabelsBreak[DESTINATION_NODE]].arrivalTime;
			breaktaken = allLabels[nodeLabelsBreak[DESTINATION_NODE]].breakTaken;
			breaknode = allLabels[nodeLabelsBreak[DESTINATION_NODE]].breakNode;
			breaktime = allLabels[nodeLabelsBreak[breaknode]].arrivalTime;
			shift_duration = allLabels[nodeLabelsBreak[DESTINATION_NODE]].shiftDuration;
			//writePath(allLabels[nodeLabelsBreak[DESTINATION_NODE]],  allLabels, results_file);
		}
	}
	if( xarrival_time != std::numeric_limits<int>::max() ) {
		xarrival_time = normalizedTime(xarrival_time);
	}
	std::ofstream results;
	results.open(results_file, std::ofstream::out | std::ofstream::app);
	//results << std::endl << TESTCASE_NUMBER <<" " << xarrival_time << " " << breaktaken << " " << breaknode << " " << breaktime << " - " <<((double)(std::clock() - start)/1000) << " sec" << std::endl;
	results << TESTCASE_NUMBER <<" " << xarrival_time << " " << breaktaken << " " << breaknode << " " <<((double)(std::clock() - start)/1000) << " " << node_errors.size() << std::endl;
	return xarrival_time;

}

};