#pragma once
#include <boost/heap/binomial_heap.hpp>
#include "Label.h"
#include "PQItem.h"
#include <vector>
#include <algorithm>
using namespace boost::heap;
using namespace std;


class SearchPQ {
	private:
		typedef boost::heap::binomial_heap<PQItem>::handle_type HeapHandle;

		vector<HeapHandle> handles;
		binomial_heap<PQItem> pq;
		binomial_heap<PQItem> pq_break;

	public:

		SearchPQ(void){}
		~SearchPQ(void){}

		void addLabel_break(int labelid, int priority){
			HeapHandle h = pq_break.push( PQItem(labelid, priority) );
			handles.push_back(h); //index == labelid!
			assert(labelid == handles.size()-1);
		}
		void addLabel_nobreak(int labelid, int priority){
			HeapHandle h = pq.push( PQItem(labelid, priority) );
			handles.push_back(h); //index == labelid!
			assert(labelid == handles.size()-1);
		}


		int getFirstLabel_nobreak(){
			return pq.top().labelid;
		}
		int getFirstLabel_break(){
			return pq_break.top().labelid;
		}

		int getPriorityFirst_nobreak(){
			return pq.top().priority;
		}
		int getPriorityFirst_break(){
			return pq_break.top().priority;
		}

		void deleteFirstLabel_nobreak(){
			pq.pop();
		}
		void deleteFirstLabel_break(){
			pq_break.pop();
		}

		void removeLabel_nobreak(int labelid){
			HeapHandle h = handles[labelid];
			pq.erase(h);
		}

		void removeLabel_break(int labelid){
			HeapHandle h = handles[labelid];
			pq_break.erase(h);
		}

		bool isEmpty_nobreak(){
			return pq.empty();
		}
		bool isEmpty_break(){
			return pq_break.empty();
		}

		void clear(){
			handles = vector<HeapHandle>();
			while(!isEmpty_break()){
				deleteFirstLabel_break();
			}
			while(!isEmpty_nobreak()){
				deleteFirstLabel_nobreak();
			}
		}
};