#ifndef RITCH_DATASTR_BLOCKING_PROFILE_H_
#define RITCH_DATASTR_BLOCKING_PROFILE_H_

#include <algorithm>
#include <cstdint>
#include <iterator>
#include <vector>

#include "ritch/util/misc.h"
#include "ritch/io/hwy_blocks_format.h"

namespace ritch
{

class BlockingProfile
{
private:

	struct Blocking
	{
		int32_t _start_time; // taken from [0, _period)
		int32_t _duration;
	};

	int32_t _period;
	std::vector<Blocking> _blockings;

public:

	BlockingProfile(const int32_t period, const std::vector<io::hwy_blocks_format::BlockingDescr>& blocking_info)
	: _period(period), _blockings()
	{
		_blockings.reserve(std::distance(blocking_info.begin(), blocking_info.end()));

		for ( auto it = blocking_info.begin() ; it != blocking_info.end() ; ++it )
		{
			int32_t duration = it->_end_time - it->_start_time;
			if ( duration < 0 ) duration += _period;

			_blockings.push_back(Blocking { it->_start_time, duration });
		}

		std::sort(_blockings.begin(), _blockings.end(),
			[](Blocking& lhs, Blocking& rhs) { return lhs._start_time < rhs._start_time; });

		if ( 
			util::has_adjacent(_blockings.begin(), _blockings.end(),
				[](Blocking& lhs, Blocking& rhs) { return lhs._start_time + lhs._duration >= rhs._start_time; })
		)
			throw std::exception("Found overlapping highway blocking.", 0);
	}

	int32_t eval_waiting_time(int32_t departure_time) const noexcept
	{
		departure_time = util::modulo(departure_time, _period);

		for ( const auto& blocking : _blockings )
		{
			if ( departure_time < blocking._start_time )
				return 0;

			if ( departure_time <= blocking._start_time + blocking._duration )
				return blocking._start_time + blocking._duration - departure_time;
		}

		return 0;
	}
};

}

#endif // RITCH_DATASTR_BLOCKING_PROFILE_H_
