#ifndef RITCH_IO_HWY_BLOCKS_FORMAT_H_
#define RITCH_IO_HWY_BLOCKS_FORMAT_H_

#include <cstdint>
#include <iostream>
#include <string>
#include <unordered_map>
#include <vector>

#include "ritch/io/binary.h"

namespace ritch
{
namespace io
{
namespace hwy_blocks_format
{

constexpr uint16_t PHONE_CODE_CZ() { return 420; };
constexpr uint16_t PHONE_CODE_SK() { return 421; };

struct BlockingDescr
{
	uint16_t _phone_code;
	int32_t _start_time;
	int32_t _end_time;

	friend void swap(BlockingDescr& lhs, BlockingDescr& rhs)
	{
		if ( &lhs != &rhs )
		{
			BlockingDescr aux = lhs;
			lhs = rhs;
			rhs = aux;
		}
	}
};

std::vector<BlockingDescr> read_blockings(const std::string& path_name)
{
	io::InputFileIterator<std::array<uint8_t, 10>> blocking_it(path_name);

	std::vector<BlockingDescr> result;
	result.resize(blocking_it.get_size_in_byte());

	std::cerr << "STATUS: Reading '" << path_name << "'... " << std::flush;
	for ( size_t pos = 0 ; ! blocking_it.finished() ; ++pos )
	{
		std::array<uint8_t, 10> blocking_item = blocking_it.read();

		result[pos]._phone_code = *reinterpret_cast<uint16_t*>(&blocking_item);
		result[pos]._start_time = *reinterpret_cast<uint32_t*>(&(blocking_item[2]));
		result[pos]._end_time = *reinterpret_cast<uint32_t*>(&(blocking_item[6]));
	}

	std::cerr << "OK, adapting... " << std::flush;
	
	for ( auto& descr : result )
	{
		descr._start_time /= 10;
		descr._end_time /= 10;
	}

	std::cerr << "OK" << std::endl;

	return result;
}

}
}
}

#endif // RITCH_IO_HWY_BLOCKS_FORMAT_H_
