#ifndef RITCH_PREPRO_EDGE_DESCRIPTION_H_
#define RITCH_PREPRO_EDGE_DESCRIPTION_H_

#include <cstdint>

#include "ritch/prepro/dynamic_graph.h"

namespace ritch
{


class EdgeDescr
{

public:
	typedef DynamicGraph::TimeDist TimeDist;
	typedef DynamicGraph::NodeId NodeId;

	NodeId _source_id;
	NodeId _target_id;
	NodeId _middle_id;
	uint32_t _complexity;
	TimeDist _time_dist;
	uint32_t _n_original_edges;
	std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> _block;

	EdgeDescr(const EdgeDescr&) = delete;
	EdgeDescr& operator= (const EdgeDescr&) = delete;

	EdgeDescr(EdgeDescr&&) = default;
	EdgeDescr& operator= (EdgeDescr&&) = default;

	EdgeDescr()
		: _source_id(NodeId::INVALID()), _target_id(NodeId::INVALID()), _middle_id(NodeId::INVALID()),
		_complexity(0), _time_dist(), _n_original_edges(0), _block()
	{}

	EdgeDescr(NodeId source_id, NodeId target_id, NodeId middle_id, TimeDist&& time_dist, std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> block, uint32_t n_original_edges)
		: _source_id(source_id),
		_target_id(target_id),
		_middle_id(middle_id),
		_complexity(uint32_t(time_dist.get_time_complexity())),
		_time_dist(std::move(time_dist)),
		_block(block),
	  _n_original_edges(n_original_edges)
	{}

	EdgeDescr(NodeId source_id, NodeId target_id, NodeId middle_id, uint32_t complexity, uint32_t n_original_edges)
	: _source_id(source_id),
	  _target_id(target_id),
	  _middle_id(middle_id),
	  _complexity(complexity),
	  _time_dist(),
	  _block(),
	  _n_original_edges(n_original_edges)
	{}

	friend std::ostream& operator<< (std::ostream& os, const EdgeDescr& descr)
	{
		os
			<< "{src=" << descr._source_id
			<< ", tgt=" << descr._target_id
			<< ", middle=" << descr._middle_id
			<< ", complexity=" << descr._complexity
			<< ", time_dist=" << descr._time_dist
			<< ", n_original=" << descr._n_original_edges
			<< "}";

		return os;
	}
};

}

#endif // RITCH_PREPRO_EDGE_DESCRIPTION_H_
