#ifndef RITCH_QUERYING_LOCATION_H_
#define RITCH_QUERYING_LOCATION_H_

#include <vector>
#include <tuple>

template <typename NodeIdType, typename TimeDistType>
class Location_tmpl
{
public:
	typedef NodeIdType NodeId;
	typedef TimeDistType TimeDist;

private:
	std::vector<std::tuple<NodeId, TimeDist>> _entrance_nodes;

public:
	Location_tmpl()
	: _entrance_nodes()
	{}

	Location_tmpl(NodeId node_id)
	: _entrance_nodes(1, std::make_pair(node_id, TimeDist(0, 0)))
	{}

	Location_tmpl(std::initializer_list<std::tuple<NodeId, int32_t, int32_t>> constant_time_dist_entrance_nodes)
	: _entrance_nodes()
	{
		for ( const auto& item : constant_time_dist_entrance_nodes )
			_entrance_nodes.emplace_back(
				std::make_tuple(std::get<0>(item), TimeDist(std::get<1>(item), std::get<2>(item))));
	}

	Location_tmpl(std::initializer_list<std::tuple<NodeId, TimeDist>> constant_time_entrance_nodes)
	: _entrance_nodes()
	{
		for ( const auto& item : constant_time_entrance_nodes )
			_entrance_nodes.emplace_back(
				std::make_tuple(std::get<0>(item), std::get<1>(item)));
	}

	size_t size() const noexcept
	{
		return _entrance_nodes.size();
	}

	const NodeId& get_node_id(size_t index) const noexcept
	{
		return std::get<0>(_entrance_nodes[index]);
	}

	const TimeDist& get_time_dist(size_t index) const noexcept
	{
		return std::get<1>(_entrance_nodes[index]);
	}
};


#endif
