#ifndef RITCH_UTIL_SHALLOW_WEAK_COPY_REF_H_
#define RITCH_UTIL_SHALLOW_WEAK_COPY_REF_H_

namespace ritch
{
namespace util
{

template <typename T>
class ShallowWeakCopyRef
{
private:
	std::array<uint8_t, sizeof(T)> _data;

public:
	ShallowWeakCopyRef(const ShallowWeakCopyRef&) = default;
	ShallowWeakCopyRef(ShallowWeakCopyRef&&) = default;

	ShallowWeakCopyRef() : _data() { _data.fill(0); }

	ShallowWeakCopyRef(T& object)
	: _data(reinterpret_cast<const std::array<uint8_t, sizeof(T)>&>(object))
	{}

	~ShallowWeakCopyRef() { _data.fill(0); }

	ShallowWeakCopyRef& operator= (const ShallowWeakCopyRef&) = default;
	ShallowWeakCopyRef& operator= (ShallowWeakCopyRef&&) = default;

	ShallowWeakCopyRef& operator= (T& obeject)
	{
		_data = reinterpret_cast<std::array<uint8_t, sizeof(T)>&>(object);
	}

	T& deref() { return reinterpret_cast<T&>(_data); }
	const T& deref() const { return reinterpret_cast<const T&>(_data); }
};

}
}

#endif // RITCH_UTIL_SHALLOW_WEAK_COPY_REF_H_
